/*
    WOD2CSV - Convert an AO-40 RUDAK WOD (Whole Orbit Data) file
    into a comma-delimited file suitable for import into a spreadsheet.
    
    Copyright 2001, Paul Williamson, KB5MU.
    

A WOD file consists of a header and a block of data.

WOD Header format:
    4 bytes     time T0 of the first sample in DOS time_t format
    4 bytes     end time in DOS time_t format
    2 bytes     collection interval DELTA_T in seconds
    1 byte      number of channels N in the channel list
    N bytes     list of N channel numbers
    
WOD data block format
    2*N bytes   N 2-byte values, corresponding to the channels in the order
                listed, as of time T0
    2*N bytes   N 2-byte values, same channels, as of T0 + DELTA_T
    2*N bytes   N 2-byte values, same channels, as of T0 + 2*DELTA_T
    etc.
    
Each multibyte value is in Intel byte order (most significant byte last).
    
Output format:

A single header line, containing the channel names as double-quoted strings,
separated by commas.
One line for each sampling time, consisting of decimal values in engineering
units, separated by commas.


Channel name and conversion formula info come from the file RUDCOEFX.
    

$Header: C:/Projects/RUDAKW~1/RCS/wod2csv.c,v 1.3 2001/05/24 01:00:22 ptw Exp ptw $

2001/05/17 ptw 0.01 Initial pre-release.

2001/05/23 ptw 0.02 Added byte-order portability (I hope). Also fixed a bug
                    that would make the last channel a copy of the next-to-last.

2001/05/23 ptw 0.03 Added header lines containing the input file name, start
                    and end times, column units, and channel numbers.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include "coefx.h"

#define _GNU_SOURCE
#ifndef __STDC__
#define __STDC__ 1
#include "getopt.h"
#undef  __STDC__
#else
#include "getopt.h"
#endif


/* Program version number */
#define WOD2CSV_VERNO   0.03


/* Structure containing the contents of a WOD file header. */
typedef struct {
    time_t          start_time;
    time_t          end_time;
    unsigned int    interval;
    unsigned char   num_chans;
    unsigned char   chan_list[256];
} wod_type;

static wod_type wod;


#define MAXARG  200         /* Maximum characters in a command-line argument */
#define MAXSS   10          /* Maximum number of subsystems we'll store */

/* Default for the calibration file name */
static char     calfile_name[MAXARG] = "RUDCOEFX";

/* List of desired subsystems */
static char     subsystems[MAXARG][MAXSS];
static char     *ss_list[MAXSS];

/* Input file name */
static char     infile_name[MAXARG];

/* Output file name */
static char     outfile_name[MAXARG];


/* Working flags set by the command line */
static int      check_length = 0;


/**************************************************************************
    version
    
    Prints a brief message detailing program version info
**************************************************************************/
void version(void)
{
    printf("WOD2CSV Whole Orbit Data to Comma Separated Format converter %1.2f\n"
           "Copyright 2001, Paul Williamson, KB5MU.\n\n", WOD2CSV_VERNO);
}

/**************************************************************************
    usage
    
    Prints a help message detailing command line usage.
**************************************************************************/
void usage(void)
{

    printf("wod2csv <flags> <input_file_name>\n"
           "   flags:\n"
           "      -s <subsystem_name>\n"
           "      --subsystem <subsystem_name>\n"
           "\n"
           "      -c <coefficient_file_name> (default \"RUDCOEFX\")\n"
           "      --coefficient_file <coefficient_file_name>\n"
           "\n"
           "      -o <output_file_name>      (defaults to stdout)\n"
           "      --output <output_file_name>\n"
           "\n"
           "      -l                         (asks program to check the file length)\n"
           "      --check_length\n"
           "\n"
           "      -v or --version            (requests program version info)\n"
           "      -h or --help               (displays this message)\n"
           "\n"
           "    Example:\n"
           "      wod2csv -s RUDAKA -s RUDAKCANA -c RUDCOEFX -o myout.csv wd051300\n");


}/* usage() */


/**************************************************************************
    read_intel
    
    Given the number of bytes to read, read in an integer in Intel byte
    order.
**************************************************************************/

long read_intel(int bytes, FILE *infile)
{
    long    value = 0;
    int     j;

    for (j=0; j < bytes; j++)
        value += ((long)(fgetc(infile) & 0xFF)) << (8*j);

    return value;
}/* read_intel() */

/**************************************************************************
    read_wod_header
    
**************************************************************************/

int read_wod_header(FILE *infile)
{
    time_t      duration;
    long        num_points;
    long        data_bytes;
    long        total_bytes;
    long        actual_size;
    struct stat stat_buf;
    int         j;


    wod.start_time = (time_t)        read_intel(4, infile);
    wod.end_time   = (time_t)        read_intel(4, infile);
    wod.interval   = (unsigned int)  read_intel(2, infile);
    wod.num_chans  = (unsigned char) read_intel(1, infile);

    for (j=0; j < wod.num_chans; j++)
        wod.chan_list[j] = (unsigned char) read_intel(1, infile);

    if (ferror(infile)) {
        perror("Error reading WOD header");
        exit(1);
    }

    /* Do some plausibility checks first */
    if (   wod.start_time <  568000000L      /* about 1988 C.E. */
        || wod.end_time   <  wod.start_time  /* must be in the right order */
        || wod.end_time   > 4102000000L      /* about 2100 C.E. */
        )
        return 0;   /* Flag this WOD file as bogus */
    
    if (check_length) {
        /* Find out how long the file is */
        if (fstat(fileno(infile), &stat_buf) != 0) {
            fprintf(stderr, "Error checking WOD file length: ");
            perror("");
            exit(1);
        }
        actual_size = stat_buf.st_size;
    
        /* Compute the expected size of the file */
        duration = wod.end_time - wod.start_time;
        num_points = duration / wod.interval;
        data_bytes = num_points * wod.num_chans * 2;
        total_bytes = 4 + 4 + 2 + 1 + wod.num_chans*2 + data_bytes;
    
        if (actual_size != total_bytes) {
            fprintf(stderr, "WOD length error: expected %ld, found %ld.\n",
                    total_bytes, actual_size);
            return 0;       /* Flag this WOD file as bogus */
        }
    }/*check_length*/

}/* read_wod_header() */


/*************************************************************************
    output_csv_header
    
    Given an open handle to an output file, look at the channel list
    from the header of the WOD file and output header lines:
    
    comma-delimited, double-quoted channel units names
    
    comma-delimited, decimal channel numbers
    
    comma-delimited, double-quoted channel names to serve as column
    headers in the spreadsheet.
*************************************************************************/    
void output_csv_header(FILE *outfile, char *infile_name)
{
    unsigned int j;
    char    start[30], end[30];     /* room for ctime output */

    fprintf(outfile, "\"Converted from file %s by WOD2CSV %1.2f\"\n",
                        infile_name, WOD2CSV_VERNO);

    strcpy(start, ctime(&wod.start_time));
    start[24] = '\0';
    fprintf(outfile, "\"Start time:\",%ld,%s UTC\n", wod.start_time, start);

    strcpy(end, ctime(&wod.end_time));
    end[24] = '\0';
    fprintf(outfile, "\"End time:\",%ld,%s UTC\n", wod.end_time, end);

    fprintf(outfile, ",");
    for (j=0; j+1 < wod.num_chans; j++) {
        fprintf(outfile, "\"%s\",", eng_units(wod.chan_list[j]));
    }
    fprintf(outfile, "\"%s\"\n", eng_units(wod.chan_list[j]));

    fprintf(outfile, ",");
    for (j=0; j+1 < wod.num_chans; j++) {
        fprintf(outfile, "%d,", wod.chan_list[j]);
    }
    fprintf(outfile, "%d\n", wod.chan_list[j]);

    fprintf(outfile, "\"Timestamp\",");
    for (j=0; j+1 < wod.num_chans; j++) {
        fprintf(outfile, "\"%s\",", channel_name(wod.chan_list[j]));
    }
    fprintf(outfile, "\"%s\"\n", channel_name(wod.chan_list[j]));

}/* output_csv_header() */


/************************************************************************
    output_csv_line
    
    Given:
        an open handle to an output file
        a buffer of WOD data
        the timestamp associated with the WOD data
        
    output to the file a comma-delimited line containing the timestamp
    and each of the data values in engineering units.
************************************************************************/    
void output_csv_line(FILE *outfile, unsigned int buf[], time_t timestamp)
{
    unsigned int    j;

    fprintf(outfile, "%ld,", timestamp);

    for (j=0; j+1 < wod.num_chans; j++) {
        fprintf(outfile, "%lf,", counts_to_eng_units(buf[j], wod.chan_list[j]));
    }
    fprintf(outfile, "%lf\n", counts_to_eng_units(buf[j], wod.chan_list[j]));

}/* output_csv_line() */


int main(int argc, char **argv)
{
    time_t          timestamp;
    FILE            *infile, *outfile;
    unsigned int    buf[256];         /* Room for a maximum-sized WOD line */
    int             override_output_filename = 0;
    int             num_ss_seen = 0;
    int             j;

    int             optchr;
    struct option   longopts[] = {
        { "coefficient_file", 1, NULL, 'c' },
        { "output",           1, NULL, 'o' },
        { "check_length",     0, NULL, 'l' },
        { "subsystem",        1, NULL, 's' },
        { "version",          0, NULL, 'v' },
        { "help",             0, NULL, 'h' },
    };

    if (argc <= 1) {
        version();
        usage();
        exit(0);
    }

    optind = 0;
    opterr = 0;
    while ((optchr = getopt_long(argc, argv, "c:hlo:s:v", longopts, NULL)) != -1) {
        switch (optchr) {
        case 'c':       /* Set the name of the coefficient file */
            strncpy(calfile_name, optarg, MAXARG);
            break;

        case 'h':       /* Display program help */
            usage();
            exit(0);
            break;

        case 'l':       /* Set a flag to check the file's length */
            check_length = 1;
            break;

        case 'o':       /* Override the output file name */
            strncpy(outfile_name, optarg, MAXARG);
            override_output_filename = 1;
            break;

        case 's':       /* Add a subsystem name to the approved list */
            if (num_ss_seen < MAXSS-1) {
                strncpy(subsystems[num_ss_seen], optarg, MAXARG);
                ss_list[num_ss_seen] = &subsystems[num_ss_seen][0];
                num_ss_seen++;
            }
            else {
                fprintf(stderr, "Too many subsystems specified.\n");
                usage();
                exit(1);
            }
            break;

        case 'v':       /* Print version information */
            version();
            printf("--help for usage information\n");
            exit(0);
            break;

        case '?':       /* Unknown option (as encoded by getopt) */
            fprintf(stderr, "Unknown option %c\n", optopt);
            usage();
            exit(1);
            break;

        case ':':       /* Missing argument (as encoded by getopt) */
            fprintf(stderr, "Option %c requires an argument.\n", optopt);
            usage();
            exit(1);
            break;
        }/*switch*/
    }/*while*/

    if (num_ss_seen == 0) {
        fprintf(stderr, "No subsystem names specified.\n");
        usage();
        exit(1);
    }
    ss_list[num_ss_seen] = NULL;

    if (optind >= argc) {
        fprintf(stderr, "No input file specified.\n");
        usage();
        exit(1);
    }
    if (optind < argc-1) {
        fprintf(stderr, "Only one input file may be specified.\n");
        usage();
        exit(1);
    }

    read_calibration_file(calfile_name, ss_list);

    strncpy(infile_name, argv[optind], MAXARG);

    if ((infile = fopen(infile_name, "rb")) == NULL) {
        fprintf(stderr, "Error opening input file %s: ", infile_name);
        perror("");
        exit(1);
    }

    if (override_output_filename) {
        if ((outfile = fopen(outfile_name, "w")) == NULL) {
            fprintf(stderr, "Error opening output file %s: ", outfile_name);
            perror("");
            exit(1);
        }
    }
    else
        outfile = stdout;

    if (!read_wod_header(infile)) {
        fprintf(stderr, "File %s does not appear to be a WOD file.\n",
                              infile_name);
        exit(1);
    }

    output_csv_header(outfile, infile_name);

    timestamp = wod.start_time;
    do {
        for (j=0; j < wod.num_chans; j++)
            buf[j] = (unsigned int) read_intel(2, infile);

        if (!feof(infile) && !ferror(infile))
            output_csv_line(outfile, buf, timestamp);

        timestamp += wod.interval;
    } while (!feof(infile) && !ferror(infile));

    if (ferror(infile)) {
        fprintf(stderr, "Error reading input file %s.\n", infile_name);
        exit(1);
    }

    if (ferror(outfile) || fclose(outfile)) {
        fprintf(stderr, "Error writing output file %s.\n", outfile_name);
        exit(1);
    }

    return 0;
}

